#ifndef INDEX_H
#define INDEX_H

#include "common.h"
#include <fstream>
#include <sys/time.h>

#include <siftHess/sift.h>
#include <siftHess/imgfeatures.h>
#include <siftHess/utils.h>
#include <siftHess/kdtree.h>
#include <siftHess/xform.h>


#include <hdf5.h>
#include <iostream>
#include <typeinfo>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/flann/flann.hpp>

#include <boost/algorithm/string.hpp>

#include "ioutils.h"


using namespace std;
using namespace fs;
using namespace Hess;
using namespace cv;

// Clustering Indexing
namespace cvflannIndex {

    struct timeval time_start,time_end;



    bool IsHDF5(fs::path pathString) {
        return (pathString.extension() == ".hdf5");
    }

    bool IsPGM(fs::path pathString) {
        return (pathString.extension() == ".pgm");
    }

    bool ensureExists(fs::path pathString,string description) {
        if(!fs::exists(pathString)) {
            cout << "ERROR: " << description << " " << pathString << " does not exist." << endl;
            exit(0);
        }
        return 1;
    }

    bool ensureHDF5(fs::path file, string description) {
        if(IsHDF5(file))
            return true;
        else {
            cout << "ERROR: "<< description << " " << file << " is not in .hdf5 format" << endl;
            exit(0);
        }
    }

    template<typename T>
    void save(Mat &data, fs::path file, string datasetName = "dataset") {
        if (IsHDF5(file)) {
            io::saveToFile<T>(data, file.string(), datasetName);
        }
        else {
            cout << "Cannot save to " << file << " because it does not end in .hdf5" << endl;
            exit(0);
        }
    }

    template<typename T>
    void load(Mat &data, fs::path file, string datasetName = "dataset") {
        ensureExists(file, "File");
        ensureHDF5(file, "File");
        cout << "Attempting to load data set from " << file << endl;
        io::loadFromFile<T>(data, file.string(),datasetName); // second parameter can be set to name of the dataset
        cout << "Loaded " << data.rows << " rows x " << data.cols << " columns of data" << endl;
    }

    size_t countLines(fs::path file) {
        ifstream fin(file.string().c_str());
	cout << "File opened " << endl;

        if (!fin.is_open())
            return 0;
        string line;
	cout << "Starting to count ... " << endl;
        size_t numLines = 0;
        while (getline(fin, line))
            numLines++;

	cout << "Done. Closing " << endl;
        fin.close();
	cout << "Leaving countLines " << endl;
        return numLines;
    }

    size_t countColumns(fs::path file) {
        ifstream fin(file.string().c_str());

        if (!fin.is_open())
            return 0;
        string line;

        size_t numcols = 0;
        getline(fin, line);
        istringstream s(line);
        string a;
        while(s>>a) {
            numcols++;
        }

        fin.close();
        return numcols;
    }


    template<typename T>
	void loadTxt(Mat & dataset, fs::path file)
    {
        cout << "Loading TXT data" << endl;
	    gettimeofday(&time_start,NULL);

        size_t numcols = countColumns(file);
        cout << "Determining number of rows..." << endl;
        size_t numLines = countLines(file);
        cout << file << " contains " << numLines << " rows" << endl;

        if (numLines) {

		    ifstream fin(file.string().c_str());
		    string line;

		    T *rawData = new T[numLines * numcols];
		    cout << "Allocated rawData at " << rawData << endl;

		    for (size_t curLine = 0; getline(fin, line); curLine++) {
		        istringstream s(line);

		        if (curLine % 10000 == 0)
		            cout << "Loading line " << curLine << endl;

				double value;
		        for (size_t curToken = 0; !s.eof(); curToken++) {
		            s >> value;
					rawData[(curLine * numcols) + (curToken)] = (T)value;
		        }
		    }
		    fin.close();

		    Mat data(numLines,numcols,(size_t)flann::CvType<T>().type(),rawData);
		    dataset = data;

		    gettimeofday(&time_end,NULL);
		    double elapsed = 1000000.0 * (time_end.tv_sec -time_start.tv_sec);
		    elapsed += (time_end.tv_usec - time_start.tv_usec);
            cout<<"Loading TXT data is done in "<<(double)(elapsed/60000000.0)<<" minutes"<<endl;
		}
    }

    template<typename T>
    void saveTxt(Mat &data, fs::path file) {
        cout << "Saving TXT data" << endl;

        cout << "Data contains " << data.rows << " rows and " << data.cols << " cols "<< endl;

        ofstream fout(file.string().c_str());
        string line;

        for (size_t i = 0; i < (size_t)data.rows; i++) {
	        for (size_t j = 0; j < (size_t)data.cols; j++) {
				T v = data.at<T>(i,j);
				fout<<v+0<<" "; // +0 for human readability (internally converts from T to int than char)
			}
			fout<<endl;
		}

        fout.close();
		cout<<"Data saving completed"<<endl;
    }



    template<typename T>
    int convertToHDF5(Mat &data, fs::path file) {
        if(IsHDF5(file)) {ifstream fin(file.string().c_str());
            string line;
            cout<< "The file is already in hdf5 format.\nSkipping the conversion" <<endl;
        }
        else {
            loadTxt<T>(data, file);
            cout<<"File loaded successfully"<<endl;
            save<T>(data,file.string()+".hdf5");
        }
        return 1;
    }


    void concatFiles(fs::path file1, fs::path file2, fs::path fileout) {
        ifstream fin1(file1.string().c_str());
        ifstream fin2(file2.string().c_str());
        ofstream fout(fileout.string().c_str());
        string line;
        while(getline(fin1,line)) {
            fout<<line<<endl;
        }
        while(getline(fin2,line)) {
            fout<<line<<endl;
        }
        fin1.close();
        fin2.close();
        fout.close();
    }


    /*
        The type of data it reads from file to load features or clusters should be compatible with the data type of Index class.
        For instance if the data points for features are stored as floats an integer Index will not able to read/load it.
    */

    template<typename FT,typename CT>
    class Index
    {
    public:
        Index()
        {
            index = NULL;
        }

        ~Index() {
            clusters.release();
            features.release();
            if(index!=NULL)
                delete index;
        }


        void test() {
/*
            cout<<countColumns("/home/gorkem/datasets/images/features.mkey")<<endl;

            cout<<sizeof(size_t)<<endl;
            cout<<sizeof(uchar)<<endl;
            cout<<sizeof(char)<<endl;
            cout<<sizeof(unsigned short)<<endl;
            cout<<sizeof(short)<<endl;
            cout<<sizeof(unsigned int)<<endl;
            cout<<sizeof(int)<<endl;
            cout<<sizeof(float)<<endl;
            cout<<sizeof(double)<<endl;
*/
            int x = 16050548;
            int y = 128;
//            Mat t(x,y,CV_8S);

            x = 1000000;
            y = 128;
  //          Mat tt(x,y,CV_32F);

/*
            for(int i=0;i<x;i++){
                for(int j=0;j<y;j++){
                    t.at<float>(i,j) = 3;
                }
            }
*/
            /*
            //mergeFeatureFiles("/home/gorkem/code/Indexing/","/home/gorkem/code/Indexing/test.features");
            //Mat m;
            //loadTxt(m,"/home/gorkem/code/Indexing/test.features");
            //save(m,"/home/gorkem/code/Indexing/test.hdf5");
            Mat m2;
            load(m2,"/home/gorkem/code/data/features_permis.mkey.hdf5");
//            cout<<m2<<endl;


            setFeatures(m2);
            buildClusters(10,10);
            cout<<"Features:"<<endl<<features<<endl;
            cout<<"Clusters:"<<endl<<clusters<<endl;
            */
        }


        void setFeatures(Mat &m) { features = m.clone(); }
        void setClusters(Mat &m) { clusters = m.clone(); }
        Mat* getFeatures() { return &features; }
        Mat* getClusters() { return &clusters; }

        //cvflann::Index<CT>* getIndex() { return index; }
        flann::Index_<CT>* getIndex() { return index; }

        static const size_t NUM_COLUMNS = 128;


        int loadFeatures(fs::path file) {
            cout << "Loading features from " << file << endl;
            load<FT>(features, file);
            return 1;
        }

        int saveFeatures(fs::path file) {
            cout << "Saving features to " << file << endl;
            save<FT>(features, file);
            return 1;
        }

        int loadClusters(fs::path file) {
            cout << "Loading clusters from " << file << endl;
            load<CT>(clusters, file);
            return 1;
        }

        int saveClusters(fs::path file) {
            cout << "Saving clusters to " << file << endl;
            save<CT>(clusters, file);
            return 1;
        }


        int buildClusters(size_t numClusters, size_t numIterations) {

            if (features.data == NULL) {
                cout << "ERROR: " << "Clustering cannot be done since features are not loaded" <<endl;
                return 0;
            }

            // -- Do the clustering
            cout << "Performing cluster operation " << numClusters << " clusters, " << numIterations << " iterations" << endl;
		    gettimeofday(&time_start,NULL);

			CT* rawdata = new CT[(size_t)((size_t)numClusters * (size_t)NUM_COLUMNS)];
			Mat clusternew(numClusters,NUM_COLUMNS,flann::CvType<CT>().type(),rawdata);
            clusters = clusternew;

            //clusterParams.cb_index = 0.6;
            clusterParams.branching = 10;
            clusterParams.iterations = numIterations;
            clusterParams.centers_init = cvflann::CENTERS_GONZALES;

            size_t N = flann::hierarchicalClustering<FT,CT>(features,clusters,clusterParams);
            cout<<N<<" number of clusters are created"<<endl;
            if(N<=0)
                return 0;

            clusters = clusters.rowRange(Range(0,N));

            cout<<clusters.rows<<" "<<clusters.cols<<endl;

		    gettimeofday(&time_end,NULL);
		    double elapsed = 1000000.0 * (time_end.tv_sec -time_start.tv_sec);
		    elapsed += (time_end.tv_usec - time_start.tv_usec);
            cout<<"CLUSTERING IS DONE in "<<(double)(elapsed/60000000.0)<<" minutes"<<endl;

            return 1;
        }


        int buildClusterIndex(const cvflann::IndexParams& params = cvflann::KDTreeIndexParams(1)) {
            if (clusters.data == NULL) {
                cout << "ERROR: " << "Indexing cannot be done since clusters are not loaded" <<endl;
                return 0;
            }

            cout << "Indexing clusters (" << clusters.rows << " rows)" << endl;



            index = new flann::Index_<CT>(clusters, params);
            return 1;
        }


        int saveClusterIndex(fs::path file) {
            ensureHDF5(file,"Cluster index file");
            cout << "Saving cluster index to " << file << endl;
            index->save(file.string());
            return 1;
        }

        int loadClusterIndex(fs::path file) {
            if (clusters.data == NULL) {
                cout << "ERROR: " << "Indexing cannot be loaded since clusters are not loaded" <<endl;
                return 0;
            }
            if(index!=NULL)
                delete index;
            index = new flann::Index_<CT>(clusters, cvflann::SavedIndexParams(file.string()));
            cout << "Cluster index is loaded successfully from file: " << file.string() << endl;
            return 1;
        }


        void extractFeatures(fs::path dir_path) {
            ensureExists(dir_path, "Directory");

            directory_iterator end_itr;
            for(directory_iterator itr(dir_path); itr!=end_itr; ++itr) {
                if(is_directory(itr->path())) {
                    cout << "DIR: " << itr->path().filename() << endl;
                    extractFeatures(itr->path());
                }
                else {
                    string filetype = itr->path().extension();
                    boost::to_lower(filetype);
                    if(filetype == ".pgm" || filetype == ".jpg" || filetype == ".jpeg" || filetype == ".ppm" || filetype == ".png" || filetype == ".bmp" || filetype == ".tiff" || filetype == ".tif") {
                        string filename = itr->path().string();
                        string img_filename = itr->path().string();
                        string out_filename = itr->path().string() + ".key";
                        cout << "\nProcessing image file: " << filename << endl;

                        // skip if its features are already extracted
                        if(fs::exists(out_filename)) {
                             cout << "Feature file already exists... skipping!"<< endl;
                        }
                        else {
                            IplImage* img = NULL;
                            struct feature* myfeatures;

                            fprintf( stderr, "Finding SIFT features...\n" );
                            img = cvLoadImage( img_filename.c_str(), CV_LOAD_IMAGE_GRAYSCALE );
                            if( ! img )
                                cout << "WARNING: Unable to load image from " << img_filename << endl;
                            else {
                                int n = sift_features( img, &myfeatures );
                                fprintf( stderr, "Found %d features.\n", n );
                                export_features( (char*)out_filename.c_str(), myfeatures, n );

                                cvReleaseImage(&img);
                                free(myfeatures);
                            }
                        }
                    }
                }
            }
        }



        int mergeFeatureFiles(fs::path dir_path, fs::path file, bool recursive=false) {

            ensureExists(dir_path, "Directory");

            if(file.string()=="") {
                cout << "Output file is empty. Saving in default file: features.mkey" << endl;
                file = "features.mkey";
            }

            ofstream fout(file.string().c_str());

            if (!fout.is_open())
                return 0;

            mergeFeaturesInDirectory(dir_path, fout, recursive);
            fout.close();
            return 1;
        }


        int queryImage(fs::path file) {
            ensureExists(file,"Image file");
            if(!IsPGM(file)) {
                cout << "ERROR: Image file needs to be in .pgm format" << endl;
                return 0;
            }

            if(index==NULL) {
                cout << "ERROR: Index is empty" << endl;
                return 0;
            }


            IplImage* img = NULL;
            struct feature* feats;

            string img_filename = file.string();
            string out_filename = file.string() + ".key";

            fprintf( stderr, "Finding SIFT features...\n" );
            img = cvLoadImage( img_filename.c_str(), 0 );
            if( ! img ) {
                cout << "unable to load image from " << img_filename << endl;
                return EXIT_FAILURE;
            }
            else {
                int n = sift_features( img, &feats );
                fprintf( stderr, "Found %d features.\n", n );
                export_features( (char*)out_filename.c_str(), feats, n );


                if(n>0) {
                    //convert Hess::feature to cv::Mat

                    Mat query(n,NUM_COLUMNS,flann::CvType<CT>().type());
                    for(int i=0; i<n; i++) {
                        for(int j=0; j<(int)NUM_COLUMNS; j++) {
                            query.at<CT>(i,j) = (CT)feats[i].descr[j];
                        }
                    }

                    // -- Compute nearest neighbors
                    int nn = 2;
                    Mat indices(query.rows,nn,flann::CvType<int>().type());
                    Mat distances(query.rows,nn,flann::CvType<float>().type());

                    cout << "Computing nearest neighbors" << endl;
                    index->knnSearch(query, indices, distances, nn, cvflann::SearchParams(128));

                    for(int i=0; i<query.rows; i++) {
                        cout<<"indices: "<<indices.at<int>(i,0)<<" dist: "<<distances.at<float>(i,0)<< endl;
                    }
                }

                cvReleaseImage(&img);
                free(feats);
                return 1;
            }

        }

    private:

        // int cvDataType;
        Mat clusters;
        Mat features;
        //cvflann::Index<T>  *index;
        flann::Index_<CT>  *index;
        cvflann::KMeansIndexParams clusterParams;





        void mergeFeaturesInDirectory(fs::path dir_path, ofstream &fout, bool recursive=false) {
            string line;
            size_t numfeats, featdim;

            directory_iterator end_itr;
            for(directory_iterator itr(dir_path); itr!=end_itr; ++itr) {
                if(is_directory(itr->path()) && recursive) {
                    cout << "DIR: " << itr->path().filename() << endl;
                    mergeFeaturesInDirectory(itr->path(), fout, true);
                }
                else {
                    if(itr->path().extension() == ".key") {

                        ifstream fin(itr->path().string().c_str());
                        getline(fin, line);
                        istringstream s(line);
                        s >> numfeats >> featdim;

                        for (size_t i = 0; i<numfeats; i++) {
                            getline(fin, line);
                            size_t count = 0;
                            while (count < featdim) {
                                getline(fin, line);

                                istringstream ss(line);
                                string num;
                                for (size_t j = 0; !ss.eof(); j++) {
                                    ss >> num;
                                    fout << num << " ";
                                    count++;
                                }
                            }
                            fout << endl;
                        }
                        fin.close();
                    }
                }
            }
        }





        /*
        void setCvDataType() {
            int cvDataType = 0;
            if(typeid(T)==typeid(unsigned char))        cvDataType = CV_8U;
            else if(typeid(T)==typeid(char))            cvDataType = CV_8S;
            else if(typeid(T)==typeid(unsigned short))  cvDataType = CV_16U;
            else if(typeid(T)==typeid(short))           cvDataType = CV_16S;
            else if(typeid(T)==typeid(int))             cvDataType = CV_32S;
            else if(typeid(T)==typeid(float))           cvDataType = CV_32F;
            else if(typeid(T)==typeid(double))          cvDataType = CV_64F;
            cout<<"cvDataType is set to: "<<cvDataType<<endl;
        }
*/

    };


}
#endif // INDEX_H
