#include "camera.h"

Camera::Camera(string devName) {
    v4l_convert_buf = NULL;
    memoryMap = NULL;
    vbuf = NULL;
    vmmap = NULL;

    deviceName = devName;

    // accessing the device...
    if ((dev = open (deviceName.c_str(), O_RDWR)) < 0){
        printf("Couldnt access the device\n");
        exit(-1);
    }

    // getting the video capabilities
    if (ioctl(dev, VIDIOCGCAP, conf.vcap) < 0){
        printf("Error in getting the capability...\n" );
        close (dev);
        exit(-1);
    }

    int webcamtype;
    if(sscanf(conf.vcap->name, "Philips %d webcam", &webcamtype) > -1){
        printf("This is a %s. (by video4linux)\n",conf.vcap->name);
        //printf("Model: %d\n",webcamtype );
    }


    // Loading default properties

    // channel properties
    if (ioctl (dev, VIDIOCGCHAN, conf.vchnl) == -1) {
        printf("ioctl (VIDIOCGCHAN) failed\n");
        close (dev);
        exit(-1);
    }

    // obtain the specifications of capture window
    if (ioctl (dev, VIDIOCGWIN, conf.vwin) == -1){
        printf ("Error (VIDIOCGWIN): Could not obtain capture window dimensions.\n");
        close (dev);
        exit (-1);
    }
    printf ("Capturing dimensions are : %d x %d\n", conf.vwin->width, conf.vwin->height);

    // get picture properties
    if (ioctl(dev, VIDIOCGPICT, conf.vpic) < 0){
        fprintf( stderr, "Error in getting the picture properties...\n" );
        close(dev);
        exit(-1);
    }

    printf("Picture colour :  %d\n", conf.vpic->colour);
    printf("Picture capture depth :  %d\n", conf.vpic->depth);
    printf("Picture palette :  %d\n", conf.vpic->palette);

    vbuf = new struct video_mbuf;
    if (ioctl (dev, VIDIOCGMBUF, vbuf) == -1){
        printf ("Failed to retrieve information about MMIO space.\n");
        close(dev);
        exit (-1);
    }

    // obtain memory mapped area
    memoryMap = (unsigned char*)mmap (0, vbuf->size, PROT_READ | PROT_WRITE, MAP_SHARED, dev, 0);

    if (memoryMap == NULL){
        printf ("Failed to obtain MMIO space.\n");
        close(dev);
        exit (-1);
    }

    // allocate structures
    vmmap = (struct video_mmap*)(malloc (vbuf->frames * sizeof (struct video_mmap)));

    updateVideommap();
    bufferIndex = 0;
}

Camera::~Camera() {

    // free the video_mmap structures
    if(vmmap!=NULL)
        free (vmmap);

    if(memoryMap!=NULL) {
        // unmap the capture memory
        munmap (memoryMap, vbuf->size);
    }

    if(dev>0) {
        // close the device
        close(dev);
    }

    if(vbuf!=NULL)
        delete vbuf;

    if(v4l_convert_buf!=NULL)
        delete [] v4l_convert_buf;
}



void Camera::updateVideommap() {

    for(int i=0; i < vbuf->frames; i++){
        vmmap[i].frame = i;
        vmmap[i].width = conf.Width();
        vmmap[i].height = conf.Height();
        vmmap[i].format = conf.Palette();
    }

    // request capture to each buffer except the last one
    for(int i=0; i < (vbuf->frames); i++) {
        if (ioctl (dev, VIDIOCMCAPTURE, &vmmap[i]) == -1){
            printf ("Capture request failed.\n");

        }
        else {
            if (ioctl (dev, VIDIOCSYNC, &vmmap[i]) == -1) {
                printf("sync failed\n");
                // sync request failed
            }
        }

    }
}


unsigned char* Camera::getFrame() {


    // send a request to begin capturing to the currently indexed buffer
    if (ioctl (dev, VIDIOCMCAPTURE, &vmmap[bufferIndex]) == -1) {
        // capture request failed
        printf("Capture request failed in getframe\n");
        exit(-1);
    }

    // wait for the currently indexed frame to complete capture
    if (ioctl (dev, VIDIOCSYNC, &vmmap[bufferIndex]) == -1) {
        printf("sync failed\n");
        // sync request failed
        exit(-1);
    }

    unsigned char* result = convert_yuv2rgb();

    // move bufferIndex to the next frame
    bufferIndex = bufferIndex + 1;
    if (bufferIndex == vbuf->frames) {
        // bufferIndex is indexing beyond the last buffer
        // set it to index the first buffer
        bufferIndex = 0;
    }

    // return the address of the frame data for the current buffer index
    //return memoryMap + vbuf->offsets[bufferIndex-1];
    return result;
}



unsigned char* Camera::convert_yuv2rgb() {
    int width = conf.vwin->width;
    int height = conf.vwin->height;


    if(v4l_convert_buf!=NULL) {
        delete[] v4l_convert_buf;
        v4l_convert_buf = NULL;
    }

    v4l_convert_buf = new unsigned char[width*height*3]; /* Used for converting yuv to bgr */

    unsigned char *y, *u, *v, *r, *g, *b;
    int i,tmp;

    // In order to see in real colors correct this as:
    // r = v4l_convert_buf+2;
    // g = v4l_convert_buf+1;
    // b = v4l_convert_buf+1;
    // However, then to use the openCV conversion (RGB2HSV) red and blue colors should change place
    r = v4l_convert_buf+0;
    g = v4l_convert_buf+1;
    b = v4l_convert_buf+2;

    y = memoryMap + vbuf->offsets[bufferIndex];
    u = y + (width * height);
    v = y + (width * height) + (width * height)/4;
    for (i=0; i<width*height; i++) {
        tmp=(((int)*y-16)*76284 + 104595*((int)*u-128)) >>16;
        if (tmp<0)
            tmp=0;
        if (tmp>255)
            tmp=255;
        *b=tmp;


        tmp=(((int)*y-16)*76284 - 53281*((int)*u-128) - 25625*((int)*v-128)) >>16;
        if (tmp<0)
            tmp=0;
        if (tmp>255)
            tmp=255;
        *g=tmp;

        //tmp=(((int)*y-16)*298 + 409*((int)*v-128)) >>8;
        tmp=(((int)*y-16)*76284 + 132252*((int)*v-128)) >>16;
        if (tmp<0)
            tmp=0;
        if (tmp>255)
            tmp=255;
        *r=tmp;

        if (i & 1) {
            u++;
            v++;
        }

        if (i%(width*2)==width-1) {
            u-=width/2;
            v-=width/2;
        }

        r+=3;
        g+=3;
        b+=3;
        y++;
    }
    return v4l_convert_buf;
}


int Camera::saveImagePPM(unsigned char *myimage, string filename) {

    unsigned char *img = myimage;

    int w = conf.vwin->width;
    int h = conf.vwin->height;

    FILE * fp;
    if((fp = fopen(filename.c_str(),"w")) == NULL) {
        printf("Could not open file %s for writing.\n", filename.c_str());
        return -1;
    }

    int n = w*h;

    fprintf(fp, "P6\n%d %d\n255\n", w, h);
    for(int i=0; i<n; ++i) {
        putc(*img, fp);
        img++;
        putc(*img, fp);
        img++;
        putc(*img, fp);
        img++;
    }

    fflush(fp);
    fclose(fp);
    return 0;
}


void Camera::updatePictureSettings() {
    if (ioctl(dev, VIDIOCSPICT, conf.vpic) < 0){
        printf("ERROR: Couldnt set desired picture properties...\n" );
        close(dev);
        exit(-1);
    }
    // retrieve the actual picture properties
    if (ioctl(dev, VIDIOCGPICT, conf.vpic) < 0){
        printf("ERROR: Couldnt get picture properties...\n" );
        close(dev);
        exit(-1);
    }
}

void Camera::updateWindowSettings() {
    // set the desired dimensions of the capturing images
    if (ioctl (dev, VIDIOCSWIN, conf.vwin) == -1){
        printf ("ERROR: Could not set capture window dimensions.\n");
        close(dev);
        exit(-1);
    }
    // retrieve the actual width and height of the capturing images
    if (ioctl (dev, VIDIOCGWIN, conf.vwin) == -1){
        printf ("ERROR: Could not obtain capture window dimensions.\n");
        close(dev);
        exit(-1);
    }
}

void Camera::updateChannelSettings() {
    // channel properties
    if (ioctl (dev, VIDIOCSCHAN, conf.vchnl) == -1) {
        printf("ERROR: Could not set channel information,(VIDIOCGCHAN) failed\n");
        close (dev);
        exit(-1);
    }
    if (ioctl (dev, VIDIOCGCHAN, conf.vchnl) == -1) {
        printf("ERROR: Could not obtaion channel information,(VIDIOCGCHAN) failed\n");
        close (dev);
        exit(-1);
    }
}



int Camera::Width() { return conf.Width(); }
int Camera::Height() { return conf.Height(); }
int Camera::Brightness() { return conf.Brightness(); }
int Camera::Hue() { return conf.Hue(); }
int Camera::Colour() { return conf.Colour(); }
int Camera::Contrast() { return conf.Contrast(); }
int Camera::Whiteness() { return conf.Whiteness(); }
int Camera::Depth() { return conf.Depth(); }
int Camera::Palette() { return conf.Palette(); }


void Camera::SetFrameSize(int w, int h) {
    conf.vwin->width = w;
    conf.vwin->height = h;
    updateWindowSettings();
    printf ("Capturing dimensions are : %d x %d\n", conf.vwin->width,conf.vwin->height);
    updateVideommap();
}

void Camera::Palette(int p) {
    conf.vpic->palette = p;
    updatePictureSettings();
    updateVideommap();
}

void Camera::Brightness(int p) {
    conf.vpic->brightness = p;
    updatePictureSettings();
}

void Camera::Hue(int p) {
    conf.vpic->hue = p;
    updatePictureSettings();
}

void Camera::Colour(int p) {
    conf.vpic->colour = p;
    updatePictureSettings();
}

void Camera::Contrast(int p) {
    conf.vpic->contrast = p;
    updatePictureSettings();
}

void Camera::Whiteness(int p) {
    conf.vpic->whiteness = p;
    updatePictureSettings();
}

void Camera::Depth(int p) {
    conf.vpic->depth = p;
    updatePictureSettings();
}




//! Returns current camera parameters
CameraConf Camera::getCameraConf() {
    return conf;
}

//! Sets camera parameters
void Camera::setCameraConf(const CameraConf& c) {
    conf = c;
}

//! Load the config file from file
int Camera::loadConf(string filename) {
    return conf.loadConf(filename);
}

//! Save the current configuration to the file
int Camera::saveConf(string filename) {
    return conf.loadConf(filename);
}
