#ifndef CAMERACONF_H
#define CAMERACONF_H

//#include "pwc-ioctl.h"
#include <linux/videodev.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <algorithm>
#include <memory.h>
#include "converter.h"

using namespace std;
using namespace converter;


//! Camera Configuration class
/*!
        Acts as the one-way interface to the configuration and parameters of the camera
        i.e. this class only provides member functions to read the parameters
*/
class CameraConf {
    friend class Camera;
protected:
    struct video_capability *vcap; // for videodev.h
    struct video_picture *vpic;
    struct video_window *vwin;
    struct video_channel *vchnl;

public:
    //! Default constructor
    /*!
        Loads the default parameters
    */
    CameraConf() {
        vchnl	= new struct video_channel;
        vchnl->type = 3;
        vcap	= new struct video_capability;
        vwin	= new struct video_window;
        vpic	= new struct video_picture;
    }

    CameraConf(const CameraConf& c) {
        vchnl = NULL;
        vcap = NULL;
        vwin = NULL;
        vpic = NULL;
        *this = c;
    }

    ~CameraConf() {
        delete vchnl;
        delete vcap;
        delete vwin;
        delete vpic;
    }


    //! Load the config file from file
    int loadConf(string filename) {
        if (filename=="") {
            cerr<<"Filename is empty"<<endl;
            return -1;
        }

        ifstream myfile;
        //cout<<"Filename: "<<filename.c_str()<<endl;
        myfile.open(filename.c_str(),ios::in);
        if (myfile.is_open()) {
            string linestr;
            string param, equalsign, value;
            size_t found;

            while(getline(myfile,linestr)) {

                // Capture the line before the comments
                found = linestr.find_first_of("#");
                if(found!=string::npos) {
                    linestr = linestr.substr(0,found);
                }
                // Ignore the line if it is empty
                if(isEmpty(linestr))
                    continue;

                //cout<<linestr<<endl;

                istringstream line(linestr);
                line >> param >> equalsign >> value;

                // Now check for correct parameter
                if (!param.compare("width")) {
                    vwin->width = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else if (!param.compare("height")) {
                    vwin->height = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else if (!param.compare("brightness")) {
                    vpic->brightness = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else if (!param.compare("hue")) {
                    vpic->hue = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else if (!param.compare("colour")) {
                    vpic->colour = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else if (!param.compare("contrast")) {
                    vpic->contrast = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else if (!param.compare("whiteness")) {
                    vpic->whiteness = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else if (!param.compare("depth")) {
                    vpic->depth = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else if (!param.compare("palette")) {
                    vpic->palette = str2int(value);
                    cout<<param<<" = "<<value<<endl;
                }
                else {
                    cout<<"Wrong parameter! Ignoring.."<<endl;
                }
            }
            myfile.close();
            return 0;
        }
        else {
            cerr<<"Unable to open file to read: "<<filename<<endl;
            return -1;
        }
    }

    //! Save the config file to file
    int saveConf(string filename) {
        if (filename=="") {
            cerr<<"Filename is empty. Saving into default file: camera.conf"<<endl;
            filename = "camera.conf";
        }

        ofstream myfile;
        myfile.open(filename.c_str(),ios::out);
        if (myfile.is_open()) {

            time_t t;
            time(&t);
            struct tm * timeinfo = localtime(&t);
            char buffer[100];
            strftime(buffer,100,"%X %A %B %d %Y",timeinfo);
            myfile<<"# Camera configuration file created at "<<buffer<<endl;
            myfile<<"width = "<<vwin->width<<endl;
            myfile<<"height = "<<vwin->height<<endl;
            myfile<<"brightness = "<<vpic->brightness<<endl;
            myfile<<"hue = "<<vpic->hue<<endl;
            myfile<<"colour = "<<vpic->colour<<endl;
            myfile<<"contrast = "<<vpic->contrast<<endl;
            myfile<<"whiteness = "<<vpic->whiteness<<endl;
            myfile<<"depth = "<<vpic->depth<<endl;
            myfile<<"palette = "<<vpic->palette<<endl;
            myfile.close();
            return 0;
        }
        else {
            cerr<<"Unable to open file to write: "<<filename<<endl;
            return -1;
        }
    }

    int Width()         {   return vwin->width; }
    int Height()        {   return vwin->height; }
    int Brightness()    {   return vpic->brightness; }
    int Hue()           {   return vpic->hue; }
    int Colour()        {   return vpic->colour; }
    int Contrast()      {   return vpic->contrast; }
    int Whiteness()     {   return vpic->whiteness; }
    int Depth()         {   return vpic->depth; }
    int Palette()       {   return vpic->palette; }


    //! Assignment operator
    CameraConf& operator=(const CameraConf  &rhs) {
        // Only do assignment if RHS is a different object from this.
        if (this != &rhs) {

            // Deallocate, allocate new space, copy values...
            if(vchnl != NULL)
                delete vchnl;
            if(vcap != NULL)
                delete vcap;
            if(vwin != NULL)
                delete vwin;
            if(vpic != NULL)
                delete vpic;

            vchnl	= new struct video_channel;
            vcap	= new struct video_capability;
            vwin	= new struct video_window;
            vpic	= new struct video_picture;

            memcpy(vchnl,rhs.vchnl,sizeof(struct video_channel));
            memcpy(vcap,rhs.vcap,sizeof(struct video_capability));
            memcpy(vpic,rhs.vpic,sizeof(struct video_picture));
            memcpy(vwin,rhs.vwin,sizeof(struct video_window));

            // In addition to memcopy the video_clip linked list should be copied
            if(vwin->clipcount > 0) {
                vwin->clips = new struct video_clip;
                struct video_clip *t = vwin->clips;
                struct video_clip *d = rhs.vwin->clips;
                t->x = d->x;
                t->y = d->y;
                t->width = d->width;
                t->height = d->height;

                for(int i=1;i<vwin->clipcount;i++) {
                    t->next = new struct video_clip;
                    t = t->next;
                    d = d->next;

                    t->x = d->x;
                    t->y = d->y;
                    t->width = d->width;
                    t->height = d->height;
                }
            }
        }
        return *this;
    }
};


#endif // CAMERACONF_H
