#ifndef FEATURE_H
#define FEATURE_H

#include "point.h"

using namespace std;

namespace feat {

// CLASS DEFINITION
//! Abstract class for feature properties
class FeatureProperty
{
public:
    FeatureProperty() {}
    virtual ~FeatureProperty() {}
    virtual void Print(std::ostream&) const = 0;
    //virtual bool operator==(const FeatureProperty &p) const;
};
static inline std::ostream& operator<<(std::ostream& stream, const FeatureProperty &p) {
    p.Print(stream);
    return stream;
}




// CLASS DEFINITION
//! Abstract class for features
class Feature
{
public:
    FeatureProperty *prop;

    Feature(){
        prop = NULL;
    }
    ~Feature(){
        if(prop!=NULL) {
            delete prop;
        }
    }
/*
    void addOccurance(int imageIndex, FeatureProperty *p) {
        occurances[imageIndex].push_back(p);
    }
*/
/*
    void deleteOccurance(int imageIndex, FeatureProperty *p) {
        // if there is any occurance for that image
        if(occurances.find(imageIndex) != occurances.end()) {
            PropertyList mylist = occurances[imageIndex];
            PropertyList::iterator itr;
            // go through occurances from that image
            for(itr=mylist.begin(); itr!=mylist.end(); itr++) {
                FeatureProperty* fp = *itr;
                if(*p == *fp) {
                    delete *itr;
                    mylist.erase(itr);
                }
            }
        }
    }
*/
};


// CLASS DEFINITION
//! Properties of a SIFT feature
class SiftProp : public FeatureProperty
{
public:
    ImagePoint loc; //!< image location of the keypoint
    double scl; //!< scale of the keypoint
    double ori; //!< orientation of the keypoint
public:
    SiftProp(): FeatureProperty() {
        loc.Init(0,0);
        scl = 0.0;
        ori = 0.0;
    }
    SiftProp(double x, double y, double s, double o): FeatureProperty() {
        loc.Init(x,y);
        scl = s;
        ori = o;
    }
    //! Copy Constructor
    SiftProp(const SiftProp& rhs) {
        // Deallocate, allocate new space, copy values...
        loc = rhs.loc;
        ori = rhs.ori;
        scl = rhs.scl;
    }

    // Returns the image location of the keypoint
    ImagePoint Location() const { return loc; }
    // Returns the scale of the keypoint
    double Scale() const { return scl; }
    // Returns the orientation of the keypoint
    double Orientation() const { return ori; }


    //! Equality in siftproperties is defined based on the equality of all properties
    bool operator==(const SiftProp &rhs) const {
        return (loc==rhs.loc && scl==rhs.scl && ori==rhs.ori); //COMPARING DOUBLES!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    }

    //! Equality in siftproperties is defined based on the equality of all properties
    bool operator!=(const SiftProp &rhs) {
        return !(*this == rhs);
    }

    void Print(std::ostream &stream) const {
        stream<<" (x,y,scl,ori): "<<loc<<" "<<scl<<" "<<ori;
    }
};



// CLASS DEFINITION
//! SIFT feature class
/*!
    Holds keypoint descriptor and location, scale, and orientation of the keypoint
*/
template <class T, int D>
class SiftFeat : public Feature
{
private:
    Point<T,D> descr; //!< keypoint descriptor
    int dim; // dimension of the descriptor...
public:
    //! Default Constructor
    SiftFeat()  : Feature() {
        T data[D];
        descr.Init(data);
        dim = D;
    }

    //! Constructor
    SiftFeat(T descriptor[]) : Feature() {
        descr.Init(descriptor);
        dim = D;
    }

    //! Constructor
    SiftFeat(T descriptor[], double x, double y, double s, double o, int imgIndex) : Feature() {
        dim = D;
        descr.Init(descriptor);

        prop = new SiftProp(x,y,s,o);
    }

    //! Copy Constructor
    // I have not tested this yet!!!
    SiftFeat(const SiftFeat& rhs) : Feature() {

        // Deallocate, allocate new space, copy values...
        descr = rhs.descr;
        dim = D;

        if (prop!=NULL)
            delete prop;
        SiftProp *p = (SiftProp*)rhs.prop;
        prop = new SiftProp(p->Location().X(), p->Location().Y(), p->Scale(), p->Orientation());

    }

    //! Dimension of its descriptor
    int Dim() { return dim; }

    Point<T,D> Descr() { return descr; }
    void setDescr(Point<T,D> data) {
        descr = data;
    }


    SiftFeat<T,D>& operator=(const SiftFeat<T,D> &rhs) {
        // Only do assignment if RHS is a different object from this.
        if (this != &rhs) {
            // Deallocate, allocate new space, copy values...
            descr = rhs.descr;

            if (prop!=NULL)
                delete prop;
            SiftProp *p = (SiftProp*)rhs.prop;
            prop = new SiftProp(p->Location().X(), p->Location().Y(), p->Scale(), p->Orientation());
        }
        return *this;
    }

    //! Equality in siftfeatures is defined based on the equality of the descriptors
    bool operator==(const SiftFeat &rhs) {
        return (descr == rhs.descr);
    }

    //! Equality in siftfeatures is defined based on the equality of the descriptors
    bool operator!=(const SiftFeat &rhs) {
        return !(descr == rhs.descr);
    }

    friend std::ostream& operator<<(std::ostream& stream, const SiftFeat &p) {
        stream<<"Feature properties:"<<endl;
        stream << *p.prop << endl;
        stream<<"Descriptor:"<<endl;
        stream<<p.descr;
        return stream;
    }
};


typedef SiftFeat<int,128> Sift128INT;
typedef SiftFeat<double,128> Sift128DOUBLE;
typedef SiftFeat<float,128> Sift128FLOAT;
typedef SiftFeat<int,64> Sift64INT;
typedef SiftFeat<double,64> Sift64DOUBLE;

typedef Sift128DOUBLE Feat;
typedef SiftProp FeatProp;

}

#endif // FEATURE_H
