#ifndef GRAPH_H
#define GRAPH_H

#include <list>
#include <set>
#include <vector>
#include <map>
#include <limits>
#include <boost/graph/adjacency_list.hpp>
#include <boost/unordered_map.hpp>
#include "opencv/cv.h"
//#include <opencv2/core/core.hpp>
#include "defs.h"

using namespace std;
using namespace boost;

class Image;


struct DispPose {
    cv::Point2d center;
    double theta;
    int boxWidth;
    int boxHeight;

    DispPose() {
        theta = 0.0;
        boxWidth = 0;
        boxHeight = 0;
        center = cv::Point2d(0.0,0.0);
    }
    bool isInside(double x, double y) {
        if(x < center.x+boxWidth/2 && x > center.x-boxWidth/2 &&
           y < center.y+boxHeight/2 && y > center.y-boxHeight/2 ) {
            return true;
        }
        return false;
    }
    DispPose& operator=(const DispPose& p) {
        theta = p.theta;
        boxWidth = p.boxWidth;
        boxHeight = p.boxHeight;
        center = p.center;
        return *this;
    }
};

struct Vertex {
    //! Vertex signiture
    int index;
    int oldindex;
    Image *image; // a vertex contains only 1 image
    string label;
    DispPose circPose;
    DispPose graphPose;
    DispPose gridmapPose;
    int mapId; //mostly used for merged maps, to identify which vertices belong to which map
    int clusterId;
    int customVar;

    Vertex() {
        index = -1;
        oldindex = -1;
        mapId = 0;
        clusterId = 0;
        customVar = 0;
        image = NULL;
    }

    Vertex& operator=(const Vertex& v) {
        index = v.index;
        oldindex = v.oldindex;
        label = v.label;

        mapId = v.mapId;
        clusterId = v.clusterId;
        customVar = v.customVar;

        circPose = v.circPose;
        graphPose = v.graphPose;
        gridmapPose = v.gridmapPose;

        //image = new Image(v.image); // should I copy the pointer or whole image???
        image = v.image;

        return *this;
    }
};

struct Edge {
    int v1,v2;
    int numMatches;
    double weight;
    double cost; // 1/weight
    CvMat *F; // Fundamental matrix connecting two images
    ZointMatchList *matches; // first column is for g[source(*e,g)].index and the second is g[target(*e,g)].index
    DispPose circPose;
    DispPose graphPose;
    DispPose gridmapPose;
    int customVar; // currently used to identify false edges
    int mapId;

    Edge()
    {
        v1=0;
        v2=0;
        numMatches=0;
        weight=0.0;
        cost=numeric_limits<double>::max();
        F=NULL;
        matches=NULL;
        customVar=0;
        mapId=0;
    }

    ~Edge() {
        if(F!=NULL) {
            cvReleaseMat(&F);
            F=NULL;
        }
        if(matches!=NULL) {
            delete matches;
            matches=NULL;
        }
    }

    Edge& operator=(const Edge& e) {
        v1 = e.v1;
        v2 = e.v2;
        numMatches = e.numMatches;
        weight = e.weight;
        cost = e.cost;
        F = cvCloneMat(e.F);
        circPose = e.circPose;
        graphPose = e.graphPose;
        gridmapPose = e.gridmapPose;
        customVar = e.customVar;
        mapId = e.mapId;

        matches = new ZointMatchList;
        for(unsigned int i=0; i<e.matches->first.size(); i++) {
            matches->first.push_back(e.matches->first[i]);
            matches->second.push_back(e.matches->second[i]);
        }

        return *this;
    }
};


struct edgeptr_t
{    typedef edge_property_tag kind; };
//struct vertexptr_t
//{    typedef vertex_property_tag kind; };


typedef property<edgeptr_t,Edge> EdgePtr;
typedef property<edge_weight_t,int,EdgePtr> EdgeProperty;
//typedef property<vertexptr_t,Vertex*> VertexProperty;


typedef adjacency_list<listS, vecS, undirectedS, Vertex, EdgeProperty, disallow_parallel_edge_tag> Graph;
typedef graph_traits<Graph>::vertex_descriptor vertexDescriptor;
typedef graph_traits<Graph>::vertex_iterator vertexIterator;
typedef graph_traits<Graph>::edge_descriptor edgeDescriptor;
typedef graph_traits<Graph>::edge_iterator edgeIterator;
typedef graph_traits<Graph>::in_edge_iterator inedgeIterator;
typedef graph_traits<Graph>::out_edge_iterator outedgeIterator;
typedef pair<edgeDescriptor,bool> edgeDescr;
typedef graph_traits<Graph>::adjacency_iterator adjacencyIterator;

//typedef property_map<Graph, vertex_index_t >::type IndexMap;




typedef property_map<Graph,edge_weight_t> WeightList;
typedef property_map<Graph,edgeptr_t> EdgeList;

#endif // GRAPH_H
