#include "image.h"


// For feature extraction use Hess' code which uses opencv
#include "sift.h"
#include "imgfeatures.h"
#include "utils.h"
#include "kdtree.h"
#include "xform.h"

#include "opencv/cv.h"
#include "opencv/highgui.h"

/*
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/features2d/features2d.hpp>
*/

using namespace iu;


Image::Image() {
    title = "";
    index = -1;    
    data = NULL;
    width = 0;
    height = 0;
    imageFormat = UNKNOWN;
    setFilename("");
    gettimeofday(&timestamp,NULL);
}

//! Constructor
Image::Image(unsigned char* i, int w, int h, ImageFormat f, Pose p) {
    title = "";
    index = -1;
    data = NULL;
    width = 0;
    height = 0;
    imageFormat = UNKNOWN;
    setImageData(i, w, h, f, p);
    setFilename("");
    gettimeofday(&timestamp,NULL);
}

//! Copy constructor
Image::Image(Image *img) {
    data = NULL;
    width = 0;
    height = 0;
    imageFormat = UNKNOWN;
    index = img->getIndex();
    setImageData(img->getImageData(),img->getWidth(),img->getHeight(),img->getImageFormat(),img->getPose());
    setTitle(img->getTitle());
    setFilename(img->getFilename());
    setFileFormat(img->fileFormat);
    setTimestamp(img->getTimestamp());
    setPose(img->getPose());
    vertex = img->vertex;
    features.clear();
    for(int i=0; i<(int)img->features.size(); i++) {
        Feat * myfeat = new Feat(*(img->features[i]));
        features.push_back(myfeat);
    }
    for(int i=0; i<(int)img->words.size(); i++) {
        words.push_back(img->words[i]);
    }
}

//! Destructor
Image::~Image() {
    //cout<<"image destructor"<<endl;
    deleteData();
    words.clear();

    // clear features list
    for(int i=0; i<(int)features.size(); i++) {
        delete features[i];
    }
    features.clear();
}

void Image::deleteData()
{
    //cout<<"deleteData"<<endl;
    if(data!=NULL) {
        delete[] data;
        data = NULL;
    }
    // clear features list
    //for(int i=0; i<(int)features.size(); i++) {
    //    delete features[i];
    //}
    //features.clear();

    width = 0;
    height = 0;
    imageFormat = UNKNOWN;
}

unsigned char* Image::getImageData() {
    return data;
}

void Image::setImageData(unsigned char* i, int w, int h, ImageFormat f, Pose p) {
    if (i!=NULL) {
        if (!(f == RGB || f == GRAY)) {
            printf("ERROR: Unknown image format\n");
            //exit(-1);
        }
        else {
            if(data!=NULL) {
                delete [] data;
                data = NULL;
            }
            width = w;
            height = h;
            imageFormat = f;
            int size = width*height*imageFormat;
            data = new unsigned char[size];
            memcpy(data,i,size);
            pose = p;
        }
    }
    else {
        //printf("ERROR: Empty image data is passed\n");
    }
}


//! This function saves the image into a file in a specified format
int Image::saveImage(string fname, FileFormat ff) {
    if(data==NULL) {
        printf("ERROR: Image is empty.\n");
        return -1;
    }
    FILE *fp;
    if((fp = fopen(fname.c_str(), "w")) == NULL) {
        printf("ERROR: Could not open file %s for writing.\n",fname.c_str());
        return -1;
    }
#if DEBUG > 1
    printf("Saving file: %s\n",fname.c_str());
#endif
    setFilename(fname);
    title = fileName.substr(fileName.find_last_of('/')+1); //filename (e.g. image.ppm)

    if(ff==PGM) {
        int n = width*height;
        fprintf(fp,"P5\n");
        fprintf(fp,"#pose: %lf %lf %lf\n",pose.x,pose.y,pose.z); // Z = theta
        fprintf(fp,"%d %d\n255\n",width,height);
        for(int i=0; i<n; i++) {
            int y;
            if(imageFormat==RGB) {
                // RGB to Grayscale conversion (formula taken from MATLAB documentation)
                // Y = 0.2989 * R + 0.5870 * G + 0.1140 * B
                y = data[i*3]*0.2989 + data[i*3+1]*0.5870 + data[i*3+2]*0.1140;
                putc((unsigned char)y,fp);
            }
            else if(imageFormat==GRAY) {
                y = data[i];
                putc((unsigned char)y,fp);
            }

        }
    }
    else if(ff==PPM) {
        int n = width*height;
        fprintf(fp,"P6\n");
        fprintf(fp,"#pose: %lf %lf %lf\n",pose.x,pose.y,pose.z); // Z = theta
        fprintf(fp,"%d %d\n255\n", width, height);
        for(int i=0; i<n; i++) {
            if(imageFormat==RGB) {
                putc(data[i*3+0], fp);
                putc(data[i*3+1], fp);
                putc(data[i*3+2], fp);
            }
            else if(imageFormat==GRAY) {
                putc(data[i*3+0], fp);
                putc(data[i*3+0], fp);
                putc(data[i*3+0], fp);
            }
        }
    }
    else {
        printf("ERROR: Unknown file format\n");
        return -1;
    }
    fflush(fp);
    fclose(fp);
    return 0;
}


//! Loads the image from the stored filename
int Image::loadImageFromFile() {
    if(fileName == "")
        return -1;
    loadImage(fileName,fileFormat);
    return 0;
}


//! This function loads the image into a file in a specified format
int Image::loadImage(string fname, FileFormat ff) {
    FILE *fp;
    if((fp = fopen(fname.c_str(), "r")) == NULL) {
        printf("ERROR: Could not open file %s for reading.\n",fname.c_str());
        return -1;
    }
    Debug(2)<<"Loading file: "<<fname.c_str()<<endl;

    setFilename(fname);
    title = fileName.substr(fileName.find_last_of('/')+1); //filename (e.g. image.ppm)

    int w,h;
    if(ff==PGM) {

        double x,y,theta,time;

        int len = 100;
        char header[len];
        fgets(header,len,fp); // P5

        long curpos = ftell(fp);
        fgets(header,len,fp);
        if(sscanf(header,"#pose: %lf %lf %lf",&x,&y,&theta)) {
            pose = Pose(x,y,theta);
            //cout<<"done pose: "<<x<<" "<<y<<" "<<theta<<endl;
        }
        else {
            size_t found;
            string stemp = string(header);
            found = stemp.find("GIMP");
            if(found!=string::npos) {
            }
            else {
                fseek(fp,curpos,SEEK_SET);
            }
        }

        fgets(header,len,fp);
        sscanf(header, "%d %d", &w, &h);
        fgets(header,len,fp); //255


        if(data!=NULL) {
            delete [] data;
            data = NULL;
        }
        width = w;
        height = h;
        imageFormat = GRAY;
        fileFormat = PGM;
        int n = width*height;
        data = new unsigned char[n*imageFormat];

        // Read Data
        for(int i=0; i<n; i++) {
            data[i] = getc(fp);
        }
    }
    else if(ff==PPM) {
        double x,y,theta,time;

        //fscanf(fp, "P6\n%d %d\n255\n", &w, &h);

        int len = 500;
        char header[len];
        fgets(header,len,fp); // P6

        long curpos = ftell(fp);
        fgets(header,len,fp);
        if(sscanf(header,"#pose: %lf %lf %lf",&x,&y,&theta)) {
            pose = Pose(x,y,theta);
            //cout<<"done pose: "<<x<<" "<<y<<" "<<theta<<endl;
        }
        else {
            size_t found;
            string stemp = string(header);
            found = stemp.find("GIMP");
            if(found!=string::npos) {
            }
            else {
                fseek(fp,curpos,SEEK_SET);
            }
        }

        fgets(header,len,fp);
        sscanf(header, "%d %d", &w, &h);
        fgets(header,len,fp); //255


        if(data!=NULL) {
            delete [] data;
            data = NULL;
        }
        width = w;
        height = h;
        imageFormat = RGB;
        fileFormat = PPM;
        int n = width*height;
        data = new unsigned char[n*imageFormat];


        for(int i=0; i<n; i++) {
            data[i*3+0] = getc(fp);
            data[i*3+1] = getc(fp);
            data[i*3+2] = getc(fp);
        }
    }
    else{
        printf("ERROR: Unknown image format\n");
        return -1;
    }

    fclose(fp);
    return 0;
}



string Image::getImageFormatString() {
    string f = "";
    switch(imageFormat) {
    case -1:
        f = "Unknown";
        break;
    case 1:
        f = "Grayscale";
        break;
    case 3:
        f = "RGB";
        break;
    }
    return f;
}



int Image::extractFeatures() {

    //timespec tic,toc;
    //clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&tic);


    int numFeats;

    bool imgReady = false;
    if(isEmpty()) {
        if(getFilename()!="") {
            if(loadImageFromFile()==0)
                imgReady = true;
        }
    }
    else
        imgReady = true;

    if(imgReady) {
        IplImage *imgcv;
        imgcv = cvCreateImage(cvSize(width,height),IPL_DEPTH_8U,1);
        int gaussian = 3;
        if(imageFormat==GRAY) {
            imgcv->imageData = (char*)data;
            cvSmooth(imgcv,imgcv,CV_GAUSSIAN,gaussian,gaussian);
        }
        else if(imageFormat==RGB) {
            IplImage *imgcvrgb;
            imgcvrgb = cvCreateImage(cvSize(width,height),IPL_DEPTH_8U,3);
            imgcvrgb->imageData = (char*)data;
            cvCvtColor(imgcvrgb,imgcv,CV_RGB2GRAY);
            cvSmooth(imgcv,imgcv,CV_GAUSSIAN,gaussian,gaussian);
            cvReleaseImage(&imgcvrgb);
        }

        struct Hess::feature *feats;
        int nPts = Hess::sift_features(imgcv,&feats);

        Debug(0)<<"Features extracted - num: "<<nPts<<endl;

        // clear features list
        for(int i=0; i<(int)features.size(); i++) {
            delete features[i];
        }
        features.clear();

        // ATTN: SOMETIMES THE sift_features FUNCTION RETURNS EXACT SAME FEATURES INCLUDING ITS PROPERTIES
        for(int i=0; i<nPts; i++) {
            bool same = true;
            // Check for all duplicates
            for(int j=i+1; j<nPts; j++) {
                // check the descriptor
                for(int k=0; k<feats[i].d; k++) {
                    if(feats[i].descr[k] != feats[j].descr[k]) {
                        same = false;
                        break;
                    }
                }
                // if descrs are the same check other properties
                if(same) {
                    if( !(feats[i].x == feats[j].x && feats[i].y == feats[j].y && areSame(feats[i].scl,feats[j].scl) && areSame(feats[i].ori,feats[j].ori)) ) {
                        same = false;
                    }
                }
            }
            if (same == false || i==nPts-1) {
                Feat * myfeat = new Feat(feats[i].descr,feats[i].x,feats[i].y,feats[i].scl,feats[i].ori, index);
                features.push_back(myfeat);
            }
        }

        cvReleaseImage(&imgcv);
        free(feats);
    }
    else {
        cout<<"ERROR: no image to extract features"<<endl;
        exit(-1);
    }


    //clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&toc);
    //double time = toc.tv_sec -tic.tv_sec + (toc.tv_nsec-tic.tv_nsec)/1000000000.0;
    //cout<<"Feature extraction time: "<<time<<endl;

    return numFeats;
}


/*
void Image::matchImage(Image *img, PointMatchList &pl)
{
    if(img->features.size() == 0)
        img->extractFeatures();

    if(features.size() == 0)
        extractFeatures();

    matchFeatures(features,img->features,pl,1);

}
*/


void Image::printFeatures() {
    cout<<"Image has "<<features.size()<<" features"<<endl;
    for(unsigned int i=0; i<features.size(); i++) {
        cout<<"Feature "<<i<<":"<<endl;
        cout<<*(features[i])<<endl;
        cout<<(features[i])->Dim()<<endl;
    }
}

void Image::printFeatures(string filename) {
    ofstream out;
    out.open(filename.c_str(),ios::out);
    out<<features.size()<<" "<<appConf.FEATUREDIM<<endl;
    for(unsigned int i=0; i<features.size(); i++) {
        FeatProp *p = (FeatProp*)features[i]->prop;
        out<<p->Location()<<" "<<p->Scale()<<" "<<p->Orientation()<<endl;
        out<<features[i]->Descr()<<endl;
    }
}

void Image::printWords() {
    for(unsigned int i=0; i<words.size(); i++) {
        cout<<"Word ["<<words[i].first<<"] at "<<words[i].second<<endl;
    }
}


void Image::resize()
{
    IplImage *imgcv;
    imgcv = cvCreateImage(cvSize(width,height),IPL_DEPTH_8U,1);
    imgcv->imageData = (char*)data;

    IplImage *imgdest;
    imgdest = cvCreateImage(cvSize(width/2,height/2),IPL_DEPTH_8U,1);
    cvResize(imgcv,imgdest);

    int n = width*height/4;

    FILE *fp = fopen("a.pgm", "w");

    fprintf(fp,"P5\n%d %d\n255\n",width/2,height/2);
    for(int i=0; i<n; i++) {
        int y = imgdest->imageData[i];
        putc((unsigned char)y,fp);
    }
    fclose(fp);
}



void Image::matchImage(Image *img, ZointMatchList &matchList) {

    if(features.size()==0)
        extractFeatures();
    if(img->features.size()==0)
        img->extractFeatures();

    int n1 = features.size();
    int n2 = img->features.size();

    struct Hess::feature* feat1 = new Hess::feature[n1];
    struct Hess::feature* feat2 = new Hess::feature[n2];

    for(int i=0; i<n1; i++) {
        feat1[i].x = ((SiftProp*)features[i]->prop)->loc.X();
        feat1[i].y = ((SiftProp*)features[i]->prop)->loc.Y();
        feat1[i].type = Hess::FEATURE_LOWE;
        feat1[i].d = appConf.FEATUREDIM;
        for(int j=0; j<appConf.FEATUREDIM; j++) {
            feat1[i].descr[j] = (double)features[i]->Descr().Data()[j];
        }
    }
    for(int i=0; i<n2; i++) {
        feat2[i].x = ((SiftProp*)img->features[i]->prop)->loc.X();
        feat2[i].y = ((SiftProp*)img->features[i]->prop)->loc.Y();
        feat2[i].type = Hess::FEATURE_LOWE;
        feat2[i].d = appConf.FEATUREDIM;
        for(int j=0; j<appConf.FEATUREDIM; j++) {
            feat2[i].descr[j] = (double)img->features[i]->Descr().Data()[j];
        }
    }


    // the maximum number of keypoint NN candidates to check during BBF search
    int KDTREE_BBF_MAX_NN_CHKS = 200;

    // threshold on squared ratio of distances between NN and 2nd NN
    double NN_SQ_DIST_RATIO_THR = 0.49;


    struct Hess::feature* feat;
    struct Hess::feature** nbrs;
    struct Hess::kd_node* kd_root;

    kd_root = kdtree_build( feat1, n1 );

    for(int i = 0; i < n2; i++ )
    {
        feat = feat2 + i;
        int k = kdtree_bbf_knn( kd_root, feat, 2, &nbrs, KDTREE_BBF_MAX_NN_CHKS );

        if( k == 2 )
        {
            double d0 = descr_dist_sq( feat, nbrs[0] );
            double d1 = descr_dist_sq( feat, nbrs[1] );

            if( d0 < d1 * NN_SQ_DIST_RATIO_THR )
            {
                matchList.first.push_back(cv::Point2d(nbrs[0]->x, nbrs[0]->y));
                matchList.second.push_back(cv::Point2d(feat->x, feat->y));
                feat2[i].fwd_match = nbrs[0];
            }
        }
        free( nbrs );
    }


    /*
    int n1 = features.size();
    int n2 = img->features.size();
    Mat feat1(n1,appConf.FEATUREDIM,CV_64F);
    Mat feat2(n2,appConf.FEATUREDIM,CV_64F);
    for(int i=0; i<n1; i++) {
        for(int j=0; j<appConf.FEATUREDIM; j++) {
            feat1.at<double>(i,j) = (double)features[i]->Descr().Data()[j];
        }
    }
    for(int i=0; i<n2; i++) {
        for(int j=0; j<appConf.FEATUREDIM; j++) {
            feat2.at<double> (i,j) = (double)img->features[i]->Descr().Data()[j];
        }
    }

    int nn = 2;
    Mat indices(feat2.rows,nn,CV_32S);
    Mat dists(feat2.rows,nn,CV_32F);

    cvflannIndex::Index<double,double> idx;
    idx.setCluster(feat1);
    idx.buildClusterIndex();
    idx.getIndex()->knnSearch(feat2,indices,dists,nn, cvflann::SearchParams(128));

    for(int i=0;i<n2; i++) {
        double d0 = dists.at<float>(i,0);
        double d1 = dists.at<float>(i,1);

        if( d0 < d1 * NN_SQ_DIST_RATIO_THR ) {
            matchList.first.push_back(cv::Point2d((double)feat1[indices.at<int>(i,0)].x, (double)feat1[indices.at<int>(i,0)].y));
            matchList.second.push_back(cv::Point2d((double)feat2[i].x, (double)feat2[i].y));
        }
    }
*/

}
