#ifndef POINT_H
#define POINT_H

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <string>


namespace feat {

//! Template class of a generic point in D dimensional space with coordinates of type T
template <class T,int D> class Point
{
protected:
    T *data; //!< data stored in the vector
    int dim; //!< dimension of the vector
public:
    //! Default constructor
    Point()
    {
        dim = D;
        data = new T[dim];
        for(int i=0; i<dim; i++) {
            data[i] = (T)0;
        }
    }

    //! Copy constructor
    Point(const Point<T,D>& p) {
        dim = p.dim;
        data = new T[dim];
        for(int i=0; i<dim; i++) {
            data[i] = p.data[i];
        }
    }

    //! Constructor
    /*! Create a n-dim point with the coordinates set to a[0],...,a[n-1] */
    explicit Point(T a[])
    {
        dim = D;
        data = new T[dim];
        Init(a);
    }

    //! Destructor
    ~Point()
    {
        //printf("Point destructor\n");
        if(data!=NULL) {
            delete []data;
            data = NULL;
        }
    }

    //! Returns the dimension
    int Dim() const { return dim; }
    //! Returns the data
    T * Data() const { return data; }

    //! Initialize the coordinates of the point
    void Init(T a[])
    {
        for(int i=0;i<dim;i++) {
            data[i] = a[i];
        }
    }

    //! Returns the norm of the point/vector
    double norm() {
        double total = 0;
        for(int i=0;i<dim;i++) {
            total += pow(data[i],2);
        }
        return sqrt(total);
    }

    friend std::ostream& operator<<(std::ostream& stream, const Point<T,D> &p) {
        if(p.dim>0) {
            for(int i=0;i<p.dim-1;i++) {
                stream<<p.data[i]<<" ";
            }
            stream<<p.data[p.dim-1];
        }
        else
            stream<<"Empty point";
        return stream;
    }

    Point<T,D>& operator=(const Point<T,D>  &rhs) {

        // Only do assignment if RHS is a different object from this.
        if (this != &rhs) {
            // Deallocate, allocate new space, copy values...
            if(data!=NULL) {
                delete [] data;
                data = NULL;
            }
            dim = rhs.dim;
            data = new T[dim];
            for(int i=0; i<dim; i++) {
                data[i] = rhs.data[i];
            }
        }

        return *this;
    }

    Point<T,D>& operator+=(const Point<T,D> &rhs) {
        for(int i=0; i<dim; i++) {
            data[i] += rhs.data[i];
        }
        return *this;
    }

    Point<T,D>& operator-=(const Point<T,D> &rhs) {
        for(int i=0; i<dim; i++) {
            data[i] -= rhs.data[i];
        }
        return *this;
    }

    const Point<T,D> operator+(const Point<T,D> &p) {
        Point<T,D> result;
        for(int i=0; i<dim; i++) {
            result.data[i] = data[i] + p.data[i];
        }
        return result;
    }

    const Point<T,D> operator-(const Point<T,D> &p) {
        Point<T,D> result;
        for(int i=0; i<dim; i++) {
            result.data[i] = data[i] - p.data[i];
        }
        return result;
    }

    bool operator==(const Point<T,D> &other) const {
        // Compare the values, and return a bool result.
        for(int i=0; i<dim; i++) {
            if(data[i] != other.data[i])
                return false;
        }
        return true;
    }

    bool operator!=(const Point<T,D> &other) const {
        return !(*this == other);
    }

};


//! Point in 2-dimensional space with coordinates of type T
template <class T>
class Point2d : public Point<T,2>
{
public:
    //! Default constructor
    Point2d() : Point<T,2>() {}

    //! Copy constructor
    Point2d(const Point2d& p) : Point<T,2>(p) {}
    //! Copy constructor
    Point2d(const Point<T,2>& p) : Point<T,2>(p) {}

    //! Constructor
    Point2d(T x, T y) : Point<T,2>()
    {
        this->data[0]=x;
        this->data[1]=y;
    }

    //! Returns the x coordinate of the point
    T X() const { return this->data[0]; }
    //! Returns the y coordinate of the point
    T Y() const { return this->data[1]; }

    //! Initialize the coordinates of the point
    void Init(T x, T y)
    {
        this->data[0]=x;
        this->data[1]=y;
    }

    Point2d<T>& operator=(const Point<T,2>  &rhs) {
        // Only do assignment if RHS is a different object from this.
        if (this != &rhs) {
            // Deallocate, allocate new space, copy values...
            delete [] this->data;
            this->data = NULL;
            this->data = new T[this->dim];
            for(int i=0; i<this->dim; i++) {
                this->data[i] = rhs.Data()[i];
            }
        }
        return *this;
    }

};


//! Point in 3-dimensional space with coordinates of type T
template <class T>
class Point3d : public Point<T,3>
{
public:
    //! Default constructor
    Point3d() : Point<T,3>() {}

    //! Constructor
    Point3d(T x, T y, T z) : Point<T,3>()
    {
        this->data[0]=x;
        this->data[1]=y;
        this->data[2]=z;
    }

    //! Returns the x coordinate of the point
    T X() const { return this->data[0]; }
    //! Returns the y coordinate of the point
    T Y() const { return this->data[1]; }
    //! Returns the z coordinate of the point
    T Z() const { return this->data[2]; }

    //! Initialize the coordinates of the point
    void Init(T x, T y, T z)
    {
        this->data[0]=x;
        this->data[1]=y;
        this->data[2]=z;
    }

    Point3d<T>& operator=(const Point<T,3>  &rhs) {
        // Only do assignment if RHS is a different object from this.
        if (this != &rhs) {
            // Deallocate, allocate new space, copy values...
            delete [] this->data;
            this->data = NULL;
            this->data = new T[this->dim];
            for(int i=0; i<this->dim; i++) {
                this->data[i] = rhs.Data()[i];
            }
        }
        return *this;
    }
};

typedef Point2d<double> ImagePoint;
typedef Point2d<double> Euclidean2dPoint;
typedef Point3d<double> Euclidean3dPoint;
//typedef Point3d<double> Pose; // x,y,theta



//! Euclidean distance between two points.
template <class T,int D>
inline static double d(Point<T,D> &p1,Point<T,D> &p2) {
    return sqrt(d2(p1,p2));
}

//! Squared Euclidean distance between two points.
template <class T,int D>
inline static double d2(Point<T,D> &p1,Point<T,D> &p2) {
    double total = 0;
    for(int i=0;i<p1.Dim();i++) {
        total += pow(p1.Data()[i] - p2.Data()[i],2);
    }
    return total;
}

}

#endif // POINT_H
