#ifndef SIMGRAPH_H
#define SIMGRAPH_H

#include <boost/random.hpp>
#include "dictionary.h"
#include "astar.h"
#include "image.h"
#include "graph.h"



typedef boost::unordered_map<int,vertexDescriptor> VertexTable; // <vertexIndex,vertexDescriptor>

// This map structure is used since this way it keeps it sorted wrt value
typedef map<pair<int,int>,pair<CvMat*,ZointMatchList*> > CandidateList; // <pair<numberOfMatches, machedImageIndex>

extern string colors[150];

class SimGraph
{
    friend class SimMap;
public:
    enum DispMode {
        GRAPHMODE = 0,
        GRIDMAPMODE,
        CIRCMODE,
        MATRIXMODE
    };
private:
    Graph g;
    //AStar mysearch;
    VertexTable vertexLookup; // is a map of <vertexIndex, vertexDescriptor> to link vertices with images
    WeightList::type gWeight;
    EdgeList::type gEdge;

    double falseDistThreshdold;

    // DISP RELATED
    cvflannIndex::Index<double,double> idxGraphV;
    cvflannIndex::Index<double,double> idxGraphE;
    vector<Vertex*> vertexDispTable;
    vector<Edge*> edgeDispTable;


    string dataDir;             // local path to temporary folder

    // Statistics for the graph: an element to the list will be added whenever a new vertex/edge is created
    int numConnComp;
    int minDegree;
    int maxDegree;
    double avgDegree;

    mt19937 randengine;


public:
    SimGraph();
    ~SimGraph();

    int numVertices() { return num_vertices(g); }
    int numEdges() { return num_edges(g); }

    vertexDescriptor addVertex(Image *img);
    bool addEdges(vertexDescriptor v, CandidateList &matches, ImageTable &imageLookup, ofstream *a=NULL);
    edgeDescr addEdge(int v1Index, int v2Index);
    bool removeVertex(Image *img, vertexDescriptor v);
    bool removeVertex(int v);
    bool removeEdge(int v1,int v2);
    bool removeEdges(vector<pair<int,int> > lim1, vector<pair<int,int> > lim2);


    void rebuildVertexLookup() {
        vertexLookup.clear();
        pair<vertexIterator,vertexIterator> vrange = vertices(g);
        for(vertexIterator v=vrange.first; v!=vrange.second; v++) {
            vertexLookup[g[*v].index] = *v;
        }
    }

    Vertex* getVertex(int index) {
        if(vertexLookup.count(index)){
            return &g[vertexLookup[index]];
        }
        else
            return NULL;
    }
    Vertex* getVertex(adjacencyIterator ai) {
        return &g[*ai];
    }
    Edge* getEdge(int index1,int index2) {
        if(vertexLookup.count(index1) && vertexLookup.count(index2)) {
            pair<edgeDescriptor,bool> edgebool = edge(vertexLookup[index1],vertexLookup[index2],g);
            if(edgebool.second)
                return &gEdge[edgebool.first];
            else {
                edgebool = edge(vertexLookup[index2],vertexLookup[index1],g);
                if(edgebool.second)
                    return &gEdge[edgebool.first];
            }
        }
        return NULL;
    }
    pair<adjacencyIterator, adjacencyIterator> getEdges(int index) {
        return adjacent_vertices(vertexLookup[index], g);
    }


    void setDataDir(string);

    void computeGraphLayoutCirc(int &w, int &h, double r, double dist);
    void computeGraphLayout(string filename, string layout="dot");
    int printGraph(string graphfile, string imagefile, string imageformat, string graphlayout);

    void writeGraph(string filename);
    void readGraph(string filename, ImageTable &imageLookup);
    void writeSimMatrix(string filename);

    vector<Vertex*>* getVertexPositions() { return &vertexDispTable; }
    vector<Edge*>* getEdgePositions() { return &edgeDispTable; }

    void resetVertexLabels();
    void updateDisplay(DispMode mode);
    Vertex* getClosestVertex(int x, int y, double &distance);
    Edge* getClosestEdge(int x, int y, double &distance);

    VertexTable* getVertexLookup() { return &vertexLookup; }

    int getVertexDegree(int index) {
        return degree(vertexLookup[index],g);
    }

    void updateStats();
    void writeStats(string filename);
    vector<int> numConnCompList;    // number of connected components
    vector<int> minDegreeList;      // minimum vertex degree
    vector<int> maxDegreeList;      // maximum vertex degree
    vector<double> avgDegreeList;   // average vertex degree
    vector<int> numEdgesList;
    vector<int> numVerticesList;


    int computeClusters(vector<int> &clusters, double max_centrality=-1.0);

    Mat graphLaplacian();


    map<int,vector<int> > components;
    Vertex *lastVertex;
    bool isConnected(CandidateList &candidates);
    bool sameConnectedComponent(int v1Index, int v2Index);
    void computeConnectedComponents();
    bool improvesGraph(CandidateList &candidates);
    int isPathShorterThan(int v1,int v2, int threshold);
    int graphDistance(int v1,int v2);


    // temp test function
    void test()
    {

    }
};

#endif // SIMGRAPH_H
