#ifndef SIMMAP_H
#define SIMMAP_H

#include "simgraph.h"
#include "image.h"
#include <QString>
#include <QStringList>
#include "time.h"
#include <boost/filesystem/path.hpp>
#include <boost/random.hpp>

using namespace boost::filesystem;

//typedef boost::unordered_map<int,double> ProbabilityRow;
//typedef boost::unordered_map<int, ProbabilityRow> ProbabilityMat;



struct MergeLog {
    ofstream out;
    vector<double> looptime;
    vector<int> loopcount;
    vector<double> matchtime;
    vector<int> nummatches;
    vector<int> maxvotes;
    vector<double> fundtime;
    vector<int> numedges;
    vector<double> addtime;
    vector<double> selecttime;
    vector<double> scoretime;
    vector<double> scores;
    vector<pair<int,int> > selectedVertices;
    vector<int> graphDistances;
    vector<pair<int,int> > degrees;

    int explorationScoreCaptionInterval;
    int refinementScoreCaptionInterval;
    double timeInterval; //seconds
    double logTime; //seconds
    int edgeaddition;
    int edgecount; //counting identified edges
    double startTime;
    timespec tic,toc,temp;
    double TIMECONVERT;
    string filename;

    MergeLog () {
        reset();
    }
    ~MergeLog () {
        reset();
    }

    void reset() {
        filename = "snapshots.txt";
        TIMECONVERT = 1000000000.0;

        edgeaddition = 0;
        edgecount = 0;
        startTime = 0.0;
        explorationScoreCaptionInterval = 10^8;
        refinementScoreCaptionInterval = 10^8;
        timeInterval = 0.0;
        logTime = 0.0;

        if(out.is_open())
            out.close();
        looptime.clear();
        loopcount.clear();
        matchtime.clear();
        nummatches.clear();
        maxvotes.clear();
        fundtime.clear();
        numedges.clear();
        addtime.clear();
        selecttime.clear();
        scoretime.clear();
        scores.clear();
        selectedVertices.clear();
        graphDistances.clear();
        degrees.clear();
    }

    double Time(timespec &t) {
        return t.tv_sec + (t.tv_nsec)/TIMECONVERT;
    }
    double Tic() {
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&tic);
        return Time(tic);
    }
    double Toc() {
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&toc);
        return Time(toc)-Time(tic);
    }
    double Duration() {
        clock_gettime(CLOCK_PROCESS_CPUTIME_ID,&temp);
        return Time(temp)-startTime;
    }
    void updateStartTime(double x) {
        startTime += x;
    }

    void printLog() {
        out.open(filename.c_str());
        if(out.is_open()) {
            for(int i=0; i<loopcount.size(); i++) {
                out<<fixed<<loopcount[i]<<"\t"<<looptime[i]
                        <<"\t"<<matchtime[i]<<"\t"<<nummatches[i]<<"\t"<<maxvotes[i]<<"\t"
                        <<fundtime[i]<<"\t"<<numedges[i]<<"\t"<<addtime[i]<<"\t"<<selecttime[i]<<"\t"
                        <<scoretime[i]<<"\t"<<scores[i]<<"\t"<<endl;
                //<<selectedVertices[i].first<<"\t"<<selectedVertices[i].second<<"\t"
                //<<graphDistances[i]<<"\t"<<degrees[i].first<<"\t"<<degrees[i].second<<endl;
            }
        }
        out.close();
    }
};


class SimMap
{
    int mapId;
public:
    struct MergeMethod {
        enum Type {
            BRUTEUNIFORM = 0,   // 1/n
            BRUTEDEGREEMIN,    // degree(v)/(2*(num(E))
            QUICKCONNECT,
            BRUTEORDER,        // brute force one fixed order
            UNIFORM_FEATURE,   // #feat/total(feat)
            MARKOV_EDGE,
            MARKOV_WEIGHT,
            MARKOV_WEIGHTREV,
            MARKOV_CLUSTERIN,
            MARKOV_CLUSTEROUT,
            SMART_1,
            SMART_2,
            SMART_3,
            HU_NEWTON,
            TEST
        };
    };
    struct MergeStopCriteria {
        enum Type {
            NONE = 0,
            TIME,
            PERCENTAGE
        };
    };
    struct MergeStyle {
        enum Type {
            SIMULTANEOUS = 0,
            PAIRWISE
        };
    };
    struct LoggingCriteria {
        enum Type {
            NONE = 0,
            SINGLE,
            CONTINUOUS
        };
    };

private:
    SimGraph *sg;
    Dictionary *dict;
    ImageTable imageLookup;     // is a map of <imageIndex, Image*>

    string dataDir;             // local path to temporary folder
    path filepath;
    int maxIndex;               // maximum index in the imagetable (need it to assign a unique new index)

    struct timeval time_start,time_end;

    //bool isImageDistinct(ZointMatchMap &fm);
    int getVotes(Image *img, ZointMatchMap &fm);
    bool enforceGeometricConstraints(ZointMatchMap &matches, CandidateList &candidates, bool forceSimilarImages=false);

    vertexDescriptor addImageAsVertex(Image *img, bool keepdata=false);

    int chooseIndex(int i);

    void cleanImageList();

    bool writeImageList(string filename);
    bool readImageList(string filename);

    bool saveImageToFile(Image *img, string path);

    vector<Image*> trackingList;

    static mt19937 randengine;
    static double mergeStartTime;

    static MergeLog mergeLog;

    void unionMaps(SimMap* m,map<int,int> &oldnewIndex);
    Mat computeProbMat(MergeMethod::Type method, int &startIndex);
    static void normalizeMat(Mat &A);
    static int selectStartVertex(Mat &A);
    static int selectNextVertex(Mat row);

    void randomConnect(SimMap *m1, SimMap *m2, map<int,int> &oldnewIndex1, map<int,int> &oldnewIndex2, MergeMethod::Type method, MergeStopCriteria::Type sc=MergeStopCriteria::NONE, double val=0.0, LoggingCriteria::Type logging=LoggingCriteria::NONE);
    static SimMap* randomConnect_pairwise(SimMap** maps, int numMaps, MergeMethod::Type method, MergeStopCriteria::Type sc, double val, bool logging=false);
    void randomConnect_simultaneous(SimMap** maps, map<int,int>* oldnewIndices, int numMaps, MergeMethod::Type method, MergeStopCriteria::Type sc, double val, bool logging=false);


    void quickConnect(SimMap *m1, SimMap *m2, map<int,int> &oldnewIndex1, map<int,int> &oldnewIndex2, MergeStopCriteria::Type sc=MergeStopCriteria::NONE, double val=0.0, LoggingCriteria::Type logging=LoggingCriteria::NONE);
    static SimMap* quickConnect_pairwise(SimMap** maps, int numMaps, MergeStopCriteria::Type sc, double val, bool logging=false);
    void quickConnect_simultaneous(SimMap** maps, map<int,int>* oldnewIndices, int numMaps, MergeStopCriteria::Type sc, double val, bool logging=false);

public:

    SimMap();
    SimMap(path filename);
    ~SimMap();

    SimGraph* getGraph() { return sg; }
    Dictionary* getDictionary() { return dict; }
    ImageTable* getImages() { return &imageLookup; }
    string getDataDir() { return dataDir; }

    void clearTrackingList() { trackingList.clear(); }

    bool isEmpty();

    void setFilepath(path s) { filepath = s; }
    path getFilepath() { return filepath; }
    string getFilename() { return filepath.stem(); }

    void test(QString f1,QString f2);
    static void test2(SimMap *m1, SimMap *m2);

    // returns the assigned indices of images inserted (there may be more than 1 image insertions if backtracking is on)
    vector<int> addImage(QString imagefile, int indx = -1, bool force=false, int trackingMode=0, bool addEdge=true); // read image file and create an Image, then same as addImage(Image *)
    vector<int> addImage(Image *img, bool force=false, int trackingMode=0, bool addEdge=true); // adds image to the map if it needs to be added
    int localizeImage(Image *img); // returns localized imageindex>0

    bool addEdge(int v1, int v2);
    bool removeVertex(int v);
    bool removeEdge(int v1, int v2);
    bool removeEdges(vector<pair<int,int> > lim1, vector<pair<int,int> > lim2);
    bool removeConnectedComp(int c);

    double scoreMap();

    bool writeMap(path filepath, bool writeIndvFiles=false);
    bool readMap(path filepath);

    void printGraph(QString &imagefile, QString imageformat, QString graphlayout);

    /* STATISTICS */
    void showTiming();
    void updateStats() { sg->updateStats(); }
    void writeStats(QString filepath);
    int numVertices() { return sg->numVertices(); }
    int numEdges() { return sg->numEdges(); }
    int graphMinDegree() { return sg->minDegree; }
    int graphMaxDegree() { return sg->maxDegree; }
    double graphAvgDegree() { return sg->avgDegree; }
    int graphNumConnComp() { return sg->numConnComp; }
    vector<double> timing;
    vector<double> timingImgIndex; //done
    vector<double> timingDict; //done
    vector<double> timingLoad; //done
    vector<double> timingAdd;  //done
    vector<double> timingFund; //done
    vector<double> timingSift; //done
    vector<double> timingTotal;


    // Map Merging
    static SimMap* mergeMaps(QStringList mapfiles, MergeMethod::Type method=MergeMethod::QUICKCONNECT, MergeStyle::Type st=MergeStyle::SIMULTANEOUS, MergeStopCriteria::Type sc=MergeStopCriteria::NONE, double val=0.0, int iter=1, bool logging=false);


    // temporary function to transfer data to Matlab
    void dumpGraphHeader(ofstream &out, SimMap *m1, SimMap *m2);
    void dumpGraph(ofstream &out, double time, SimMap *m1, SimMap *m2,map<int,int> &oldnewIndex);
    void dumpGraph(SimMap *m1, SimMap *m2,map<int,int> &oldnewIndex);
};


#endif // SIMMAP_H

