#ifndef IOUTILS_H
#define IOUTILS_H

#include <iostream>
#include <typeinfo>
#include <hdf5.h>

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/flann/flann.hpp>

#include <boost/algorithm/string.hpp>
#include <boost/filesystem.hpp>

#include <fstream>
#include <sys/time.h>

namespace fs = boost::filesystem;

using namespace std;
using namespace fs;
using namespace cv;

namespace io {

#define CHECK_ERROR(x,y) if ((x)<0) { cout<<y<<endl; exit(0); }

    namespace {

        template<typename T>
        hid_t getHDF5Type()
        {
            CHECK_ERROR(-1,"Unsupported type for IO operations");
        }

        template<> hid_t getHDF5Type<char>() { return H5T_NATIVE_CHAR; }
        template<> hid_t getHDF5Type<unsigned char>() { return H5T_NATIVE_UCHAR; }
        template<> hid_t getHDF5Type<short int>() { return H5T_NATIVE_SHORT; }
        template<> hid_t getHDF5Type<unsigned short int>() { return H5T_NATIVE_USHORT; }
        template<> hid_t getHDF5Type<int>() { return H5T_NATIVE_INT; }
        template<> hid_t getHDF5Type<unsigned int>() { return H5T_NATIVE_UINT; }
        template<> hid_t getHDF5Type<long>() { return H5T_NATIVE_LONG; }
        template<> hid_t getHDF5Type<unsigned long>() { return H5T_NATIVE_ULONG; }
        template<> hid_t getHDF5Type<float>() { return H5T_NATIVE_FLOAT; }
        template<> hid_t getHDF5Type<double>() { return H5T_NATIVE_DOUBLE; }
        template<> hid_t getHDF5Type<long double>() { return H5T_NATIVE_LDOUBLE; }
    }


    template<typename T>
    void saveToFile(cv::Mat& dataset, const std::string& filename, const std::string& name)
    {
        herr_t status;
        hid_t file_id = H5Fcreate(filename.c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
        CHECK_ERROR(file_id,"Error opening hdf5 file.");

        hsize_t     dimsf[2];              // dataset dimensions
        dimsf[0] = dataset.rows;
        dimsf[1] = dataset.cols;

        hid_t space_id = H5Screate_simple(2, dimsf, NULL);
        hid_t memspace_id = H5Screate_simple(2, dimsf, NULL);

        hid_t dataset_id = H5Dcreate2(file_id, name.c_str(), getHDF5Type<T>(), space_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
        CHECK_ERROR(dataset_id,"Error creating dataset in file.");

        status = H5Dwrite(dataset_id, getHDF5Type<T>(), memspace_id, space_id, H5P_DEFAULT, (T*)dataset.ptr<T>(0) );
        CHECK_ERROR(status,"Error writing to dataset");

        H5Sclose(memspace_id);
        H5Sclose(space_id);
        H5Dclose(dataset_id);
        H5Fclose(file_id);
    }

    template<typename T>
    void loadFromFile(cv::Mat &dataset,const std::string& filename, const std::string& name)
    {
        herr_t status;
        hid_t file_id = H5Fopen(filename.c_str(), H5F_ACC_RDWR, H5P_DEFAULT);
        CHECK_ERROR(file_id,"Error opening hdf5 file.");

        hid_t dataset_id = H5Dopen2(file_id, name.c_str(), H5P_DEFAULT);
        CHECK_ERROR(dataset_id,"Error opening dataset in file.");

        hid_t space_id = H5Dget_space(dataset_id);

        hsize_t dims_out[2];
        H5Sget_simple_extent_dims(space_id, dims_out, NULL);

        //dataset = flamm::Matrix<T>(new T[dims_out[0]*dims_out[1]], dims_out[0], dims_out[1]);
        int datatype = 0;
        if(typeid(T)==typeid(unsigned char))        datatype = CV_8U;
        else if(typeid(T)==typeid(char))            datatype = CV_8S;
        else if(typeid(T)==typeid(unsigned short))  datatype = CV_16U;
        else if(typeid(T)==typeid(short))           datatype = CV_16S;
        else if(typeid(T)==typeid(int))             datatype = CV_32S;
        else if(typeid(T)==typeid(float))           datatype = CV_32F;
        else if(typeid(T)==typeid(double))          datatype = CV_64F;


        T *rawData = new T[dims_out[0] * dims_out[1]];
        if (rawData==NULL) {
            cout<<"ERROR: Couldn't allocate enough memory"<<endl;
            exit(0);
        }
        cv::Mat data(dims_out[0],dims_out[1],datatype,rawData);
        dataset = data.clone();
        data.release();
        delete []rawData;

        status = H5Dread(dataset_id, getHDF5Type<T>(), H5S_ALL, H5S_ALL, H5P_DEFAULT, (T*)dataset.ptr<T>(0));
        CHECK_ERROR(status, "Error reading dataset");

        H5Sclose(space_id);
        H5Dclose(dataset_id);
        H5Fclose(file_id);
    }

    inline bool IsHDF5(fs::path pathString) {
        return (pathString.extension() == ".hdf5");
    }

    inline bool IsPGM(fs::path pathString) {
        return (pathString.extension() == ".pgm");
    }

    inline bool ensureExists(fs::path pathString,string description) {
        if(!fs::exists(pathString)) {
            cout << "ERROR: " << description << " " << pathString << " does not exist." << endl;
            exit(0);
        }
        return 1;
    }

    inline bool ensureHDF5(fs::path file, string description) {
        if(IsHDF5(file))
            return true;
        else {
            cout << "ERROR: "<< description << " " << file << " is not in .hdf5 format" << endl;
            exit(0);
        }
    }

    template<typename T>
    void save(Mat &data, fs::path file, string datasetName = "dataset") {
        cout << "Attempting to save dataset to " << file << endl;
        if (IsHDF5(file)) {
            io::saveToFile<T>(data, file.string(), datasetName);
        }
        else {
            cout << "Cannot save to " << file << " because it does not end in .hdf5" << endl;
            exit(0);
        }
    }

    template<typename T>
    void load(Mat &data, fs::path file, string datasetName = "dataset") {
        ensureExists(file, "File");
        ensureHDF5(file, "File");
        cout << "Attempting to load dataset from " << file << endl;
        io::loadFromFile<T>(data, file.string(),datasetName); // second parameter can be set to name of the dataset
        cout << "Loaded " << data.rows << " rows x " << data.cols << " columns of data" << endl;
    }

    inline size_t countLines(fs::path file) {
        ifstream fin(file.string().c_str());
        cout << "File opened " << endl;

        if (!fin.is_open())
            return 0;
        string line;
        cout << "Starting to count ... " << endl;
        size_t numLines = 0;
        while (getline(fin, line))
            numLines++;

        cout << "Done. Closing " << endl;
        fin.close();
        cout << "Leaving countLines " << endl;
        return numLines;
    }

    inline size_t countColumns(fs::path file) {
        ifstream fin(file.string().c_str());

        if (!fin.is_open())
            return 0;
        string line;

        size_t numcols = 0;
        getline(fin, line);
        istringstream s(line);
        string a;
        while(s>>a) {
            numcols++;
        }

        fin.close();
        return numcols;
    }


    template<typename T>
    void loadTxt(Mat & dataset, fs::path file)
    {
        struct timeval time_start,time_end;

        cout << "Loading TXT data" << endl;
        gettimeofday(&time_start,NULL);

        size_t numcols = countColumns(file);
        cout << "Determining number of rows..." << endl;
        size_t numLines = countLines(file);
        cout << file << " contains " << numLines << " rows" << endl;

        if (numLines) {

            ifstream fin(file.string().c_str());
            string line;

            T *rawData = new T[numLines * numcols];
            if (rawData==NULL) {
                cout<<"ERROR: Couldn't allocate enough memory"<<endl;
                exit(0);
            }
            cout << "Allocated rawData at " << rawData << endl;

            for (size_t curLine = 0; getline(fin, line); curLine++) {
                istringstream s(line);

                if (curLine % 10000 == 0)
                    cout << "Loading line " << curLine << endl;

                double value;
                for (size_t curToken = 0; !s.eof(); curToken++) {
                    s >> value;
                    rawData[(curLine * numcols) + (curToken)] = (T)value;
                }
            }
            fin.close();

            Mat data(numLines,numcols,(size_t)flann::CvType<T>().type(),rawData);
            dataset = data;

            gettimeofday(&time_end,NULL);
            double elapsed = 1000000.0 * (time_end.tv_sec -time_start.tv_sec);
            elapsed += (time_end.tv_usec - time_start.tv_usec);
            cout<<"Loading TXT data is done in "<<(double)(elapsed/60000000.0)<<" minutes"<<endl;
        }
    }

    template<typename T>
    void saveTxt(Mat &data, fs::path file) {
        cout << "Saving TXT data" << endl;

        cout << "Data contains " << data.rows << " rows and " << data.cols << " cols "<< endl;

        ofstream fout(file.string().c_str());
        string line;

        for (size_t i = 0; i < (size_t)data.rows; i++) {
            for (size_t j = 0; j < (size_t)data.cols; j++) {
                T v = data.at<T>(i,j);
                fout<<v+0<<" "; // +0 for human readability (internally converts from T to int than char)
            }
            fout<<endl;
        }

        fout.close();
        cout<<"Data saving completed"<<endl;
    }



    template<typename T>
    int convertToHDF5(Mat &data, fs::path file) {
        if(IsHDF5(file)) {ifstream fin(file.string().c_str());
            string line;
            cout<< "The file is already in hdf5 format.\nSkipping the conversion" <<endl;
        }
        else {
            loadTxt<T>(data, file);
            cout<<"File loaded successfully"<<endl;
            save<T>(data,file.string()+".hdf5");
        }
        return 1;
    }


    inline void concatFiles(fs::path file1, fs::path file2, fs::path fileout) {
        ifstream fin1(file1.string().c_str());
        ifstream fin2(file2.string().c_str());
        ofstream fout(fileout.string().c_str());
        string line;
        while(getline(fin1,line)) {
            fout<<line<<endl;
        }
        while(getline(fin2,line)) {
            fout<<line<<endl;
        }
        fin1.close();
        fin2.close();
        fout.close();
    }



}

#endif // IOUTILS_H
